.. _attic_verify:

attic verify
------------
::


    usage: attic verify [-h] [-v] [-e PATTERN] ARCHIVE [PATH [PATH ...]]
    
    Verify archive consistency
    
    positional arguments:
      ARCHIVE               archive to verity integrity of
      PATH                  paths to verify
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --verbose         verbose output
      -e PATTERN, --exclude PATTERN
                            exclude paths matching PATTERN

Description
~~~~~~~~~~~


