.. _attic_change-passphrase:

attic change-passphrase
-----------------------
::


    usage: attic change-passphrase [-h] [-v] REPOSITORY
    
    Change repository key file passphrase
    
    positional arguments:
      REPOSITORY
    
    optional arguments:
      -h, --help     show this help message and exit
      -v, --verbose  verbose output

Description
~~~~~~~~~~~


