.. _attic_extract:

attic extract
-------------
::


    usage: attic extract [-h] [-v] [-e PATTERN] [--numeric-owner]
                         ARCHIVE [PATH [PATH ...]]
    
    Extract archive contents
    
    positional arguments:
      ARCHIVE               archive to extract
      PATH                  paths to extract
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --verbose         verbose output
      -e PATTERN, --exclude PATTERN
                            exclude paths matching PATTERN
      --numeric-owner       only obey numeric user and group identifiers

Description
~~~~~~~~~~~


