.. _attic_mount:

attic mount
-----------
::


    usage: attic mount [-h] [-v] [-f] [-o OPTIONS] ARCHIVE MOUNTPOINT
    
    Mount archive as a FUSE fileystem
    
    positional arguments:
      ARCHIVE           archive to mount
      MOUNTPOINT        where to mount filesystem
    
    optional arguments:
      -h, --help        show this help message and exit
      -v, --verbose     verbose output
      -f, --foreground  stay in foreground, do not daemonize
      -o OPTIONS        Extra mount options

Description
~~~~~~~~~~~


