.. _attic_init:

attic init
----------
::


    usage: attic init [-h] [-v] [-e {none,passphrase,keyfile}] REPOSITORY
    
    Initialize an empty repository
    
    positional arguments:
      REPOSITORY            repository to create
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --verbose         verbose output
      -e {none,passphrase,keyfile}, --encryption {none,passphrase,keyfile}
                            select encryption method

Description
~~~~~~~~~~~


