.. highlight:: bash
.. |project_name| replace:: ``Attic``
.. |package_dirname| replace:: Attic-|version|
.. |package_filename| replace:: |package_dirname|.tar.gz
.. |package_url| replace:: https://pypi.python.org/packages/source/A/Attic/|package_filename|
.. |git_url| replace:: https://github.com/jborg/attic.git
.. _deduplication: https://en.wikipedia.org/wiki/Data_deduplication
.. _AES: https://en.wikipedia.org/wiki/Advanced_Encryption_Standard
.. _HMAC-SHA256: http://en.wikipedia.org/wiki/HMAC
.. _PBKDF2: https://en.wikipedia.org/wiki/PBKDF2
.. _ACL: https://en.wikipedia.org/wiki/Access_control_list
.. _github: https://github.com/jborg/attic
.. _OpenSSL: https://www.openssl.org/
.. _Python: http://www.python.org/
.. _`msgpack-python`: https://pypi.python.org/pypi/msgpack-python/
.. _homebrew: http://mxcl.github.io/homebrew/
.. _issue tracker: https://github.com/jborg/attic/issues
.. _userspace filesystems: https://en.wikipedia.org/wiki/Filesystem_in_Userspace
.. _librelist: http://librelist.com/
