.. _attic_delete:

attic delete
------------
::


    usage: attic delete [-h] [-v] ARCHIVE
    
    Delete archive
    
    positional arguments:
      ARCHIVE        archive to delete
    
    optional arguments:
      -h, --help     show this help message and exit
      -v, --verbose  verbose output

Description
~~~~~~~~~~~


