.. _attic_info:

attic info
----------
::


    usage: attic info [-h] [-v] ARCHIVE
    
    Show archive details such as disk space used
    
    positional arguments:
      ARCHIVE        archive to display information about
    
    optional arguments:
      -h, --help     show this help message and exit
      -v, --verbose  verbose output

Description
~~~~~~~~~~~


