.. _attic_list:

attic list
----------
::


    usage: attic list [-h] [-v] REPOSITORY_OR_ARCHIVE
    
    List archive or repository contents
    
    positional arguments:
      REPOSITORY_OR_ARCHIVE
                            repository/archive to list contents of
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --verbose         verbose output

Description
~~~~~~~~~~~


